(* Delphi Dialog
   Web Updater
   ===========
   Check Version on Web for Update (Thread based)
   Download file from URL and show Progress

    Dr. J. Rathlev, D-24222 Schwentinental (info(a)rathlev-home.de)

   The contents of this file may be used under the terms of the
   Mozilla Public License ("MPL") or
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   Vers. 1 - Jul. 2008
   *)

unit WebUpdateDlg;

interface

uses Winapi.Windows, System.SysUtils, System.Classes, Vcl.Graphics, Vcl.Forms,
  Vcl.Controls, Vcl.StdCtrls, Vcl.Buttons, Vcl.ExtCtrls, Vcl.ComCtrls,
  IdComponent, IdHttp;

const
  defHttpProxyPort = 80;
  LocExt = 'loc';

type
  THttpProxyParams = record
    UseProxy       : boolean;
    Server,
    Username,
    Password       : string;
    Port           : integer;
    end;

  TUpdateStatus = (usNotAvail,usAvail,usCancel,usNoServer);
  TStatusEvent = procedure (var IsIdle : boolean) of object;

  TUpdateThread = class (TThread)
  private
    FCheckUrl,
    FVersion,
    FUrl,FName     : string;
    FLevel         : integer;
    IdHTTP         : TIdHTTP;
    FCanTerminate,
    FCancelled,
    FShowFault     : boolean;
    FUpdateStatus  : TUpdateStatus;
    FOnTermCheck   : TStatusEvent;
    procedure SetTermCheck (ACallBack : TStatusEvent);
    procedure DoTermCheck;
    procedure TermCheck;
  protected
    procedure Execute; override;
  public
    constructor Create (const ACheckUrl,Version : string; const AProxyPar : THttpProxyParams;
       ALevel : integer; AOnTermCheck : TStatusEvent; AOnTerminate :TNotifyEvent; AShowFault : boolean);
    destructor Destroy; override;
    function GetLocation : string;
    procedure Cancel;
    property DownloadUrl : string read FUrl;
    property DownloadName : string read FName;
    property UpdateVersion : string read FVersion;
    property ShowFault : boolean read FShowFault;
    property UpdateStatus : TUpdateStatus read FUpdateStatus;
    property OnCheckTerminated : TStatusEvent read FOnTermCheck write SetTermCheck;
    property Terminated;
    end;

  TDownload = class(TForm)
    edURL: TLabeledEdit;
    edFilename: TLabeledEdit;
    pbDownload: TProgressBar;
    procedure FormCreate(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
  private
    { Private declarations }
    nMax      : int64;
    IdHTTP    : TIdHTTP;
    procedure StartWork(Sender: TObject; AWorkMode: TWorkMode; AWorkCount: Int64);
    procedure ShowWork(Sender: TObject; AWorkMode: TWorkMode; AWorkCount: Int64);
    procedure FinishWork(Sender: TObject; AWorkMode: TWorkMode);
  public
    { Public declarations }
    function Execute (const ATitle,URL,Filename : string; const AProxyPar : THttpProxyParams) : boolean;
  end;

function IsNewerVersion (v1,v2 : string; Level : integer) : boolean;
procedure ResetProxy (var AProxyPar : THttpProxyParams);

function ShowDownload (const ATitle,URL,Filename : string; const AProxyPar : THttpProxyParams) : boolean;

var
  Download: TDownload;

implementation

{$R *.dfm}

uses StringUtils, FileUtils, GnuGetText;

{ ---------------------------------------------------------------- }
// True: v1<v2
function IsNewerVersion (v1,v2 : string; Level : integer) : boolean;
var
  n1,n2,i : integer;
begin
  n1:=0; i:=0;                               
  while (length(v1)>0) and (i<level) do begin
    n1:=100*n1+ReadNxtInt(v1,'.',0);
    inc(i);
    end;
  while (i<level) do begin
    n1:=100*n1;
    inc(i);
    end;
  n2:=0; i:=0;
  while (length(v2)>0) and (i<level) do begin
    n2:=100*n2+ReadNxtInt(v2,'.',0);
    inc(i);
    end;
  while (i<level) do begin
    n1:=100*n1;
    inc(i);
    end;
  Result:=n1<n2;
  end;

// Reset proxy sttings
procedure ResetProxy (var AProxyPar : THttpProxyParams);
begin
  with AProxyPar do begin
    UseProxy:=false;
    Server:='';
    Username:='';
    Password:='';
    Port:=defHttpProxyPort;
    end;
  end;

{ ---------------------------------------------------------------- }
constructor TUpdateThread.Create (const ACheckUrl,Version : string; const AProxyPar : THttpProxyParams;
                                  ALevel : integer; AOnTermCheck : TStatusEvent;
                                  AOnTerminate :TNotifyEvent; AShowFault : boolean);
begin
  inherited Create (false);
  Priority:=tpLower;
//  FHandle:=CfHandle;
  FCheckUrl:=ACheckUrl;
  FVersion:=Version;
  FLevel:=ALevel;
  FUrl:=''; FName:='';
  IdHTTP:=TIdHTTP.Create(nil);
  with IdHTTP do begin
    ReadTimeout:=10000;
    with AProxyPar do begin
      ProxyParams.BasicAuthentication:=UseProxy and (length(UserName)>0) and (length(Password)>0);
      if UseProxy then ProxyParams.ProxyServer:=Server
      else ProxyParams.ProxyServer:='';
      ProxyParams.ProxyPort:=Port;
      ProxyParams.ProxyUsername:=Username;
      ProxyParams.ProxyPassword:=Password;
      end;
    end;
  FreeOnTerminate:=true;
  FCanTerminate:=true;
  FCancelled:=false;
  FShowFault:=AShowFault;
  OnCheckTerminated:=AOnTermCheck;
  OnTerminate:=AOnTerminate;
  end;

destructor TUpdateThread.Destroy;
begin
  IdHTTP.Free;
  inherited;
  end;

procedure TUpdateThread.Cancel;
begin
  FCancelled:=true;
  end;

procedure TUpdateThread.SetTermCheck (ACallBack : TStatusEvent);
begin
  FOnTermCheck:=ACallBack;
  end;

procedure TUpdateThread.DoTermCheck;
begin
  if assigned(FOnTermCheck) then FOnTermCheck(FCanTerminate) //Synchronize(TermCheck)
  else FCanTerminate:=true;
  end;

procedure TUpdateThread.TermCheck;
begin
  FOnTermCheck(FCanTerminate);
  end;

procedure TUpdateThread.Execute;
var
  s,v           : string;
begin
  FUpdateStatus:=usNoServer;
  s:='';
  if length(FCheckUrl)>0 then begin
    try s:=IdHTTP.Get(FCheckUrl); except s:=''; end;
    if length(s)>0 then begin
      s:=CrLfToChar(s,'|');
      v:=ReadNxtStr(s,'|');
      if IsNewerVersion(FVersion,v,FLevel) then FUpdateStatus:=usAvail
      else FUpdateStatus:=usNotAvail;
      end;
    end;
  FVersion:=v;
  FUrl:=ReadNxtStr(s,'|');
  FName:=ReadNxtStr(s,'|');
//  if FHandle<>0 then PostMessage (FHandle,wmUpdateAvailable,ord(UpdateAvail),0);
  if FCancelled then FUpdateStatus:=usCancel;
//  FUpdateStatus:=usAvail; // zum Testen
  repeat
    Sleep(200);
    DoTermCheck;
    until FCanTerminate;
  Terminate;
  end;

function TUpdateThread.GetLocation : string;
var
  s : string;
begin
  if length(FCheckUrl)>0 then begin
    s:=NewExt(FCheckUrl,LocExt);
    try s:=IdHTTP.Get(s); except s:=''; end;
    if length(s)>0 then begin
      s:=CrLfToChar(s,'|');
      FUrl:=ReadNxtStr(s,'|');
      FName:=ReadNxtStr(s,'|');
      end
    end;
  Result:=FUrl+FName;
  end;

{ ---------------------------------------------------------------- }
procedure TDownload.FormCreate(Sender: TObject);
begin
  TranslateComponent (self,'dialogs');
  end;

procedure TDownload.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  if assigned(IdHTTP) then IdHTTP.Free;
  end;

procedure TDownload.StartWork(Sender: TObject; AWorkMode: TWorkMode; AWorkCount: Int64);
begin
  if AWorkMode=wmRead then nMax:=AWorkCount;
  pbDownload.Position:=0;
  Application.ProcessMessages;
  end;

procedure TDownload.ShowWork(Sender: TObject; AWorkMode: TWorkMode; AWorkCount: Int64);
begin
  if (AWorkMode=wmRead) and (nMax>0) then
    pbDownload.Position:=round((AWorkCount/nMax)*100);
  Application.ProcessMessages;
  end;

procedure TDownload.FinishWork(Sender: TObject; AWorkMode: TWorkMode);
begin
  if AWorkMode=wmRead then with pbDownload do begin
    Max:=101; Position:=101; Max:=100;
    end;
  Application.ProcessMessages;
  end;


{------------------------------------------------------------------- }
function TDownload.Execute (const ATitle,URL,Filename : string; const AProxyPar : THttpProxyParams) : boolean;
var
  fs      : TFileStream;
begin
  Caption:=ATitle;
  edURL.Text:=URL;
  edFilename.Text:=Filename;
  IdHTTP:=TIdHTTP.Create(Application);
  with IdHTTP do begin
    with AProxyPar do begin
      ProxyParams.BasicAuthentication:=UseProxy and (length(UserName)>0) and (length(Password)>0);
      if UseProxy then ProxyParams.ProxyServer:=Server
      else ProxyParams.ProxyServer:='';
      ProxyParams.ProxyPort:=Port;
      ProxyParams.ProxyUsername:=Username;
      ProxyParams.ProxyPassword:=Password;
      end;
    OnWorkBegin:=StartWork;
    OnWork:=ShowWork;
    OnWorkEnd:=FinishWork;
    end;
  Result:=false;
  Show;
  BringToFront;
  try
    fs:=TFileStream.Create(Filename,fmCreate);
    try
      IdHTTP.Get(Url,fs);
      Result:=true;
    except
      end;
  finally
    fs.Free;
    end;
  Close;
  end;

{------------------------------------------------------------------- }
function ShowDownload (const ATitle,URL,Filename : string; const AProxyPar : THttpProxyParams) : boolean;
begin
  if not assigned(Download) then Download:=TDownload.Create(Application);
  Result:=Download.Execute(ATitle,URL,Filename,AProxyPar);
  FreeAndNil(Download);
  end;

end.
